@@ -1,6 +1,6 @@
 // 
 //
-//  $Id: spamass-milter.cpp,v 1.5 2002/01/31 15:34:43 greve Exp $
+//  $Id: spamass-milter.cpp,v 1.13 2002/08/14 16:15:10 dnelson Exp $
 //
 //  SpamAss-Milter 
 //    - a rather trivial SpamAssassin Sendmail Milter plugin
@@ -11,9 +11,10 @@
 //  for information about Sendmail please see
 //                        http://www.sendmail.org
 //
-//  Copyright (c) 2002 Georg C. F. Greve <greve@gnu.org>
+//  Copyright (c) 2002 Georg C. F. Greve <greve@gnu.org>,
+//   all rights maintained by FSF Europe e.V., 
+//   Villa Vogelsang, Antonienallee 1, 45279 Essen, Germany
 //
-// Add tracking of X-Spam-Checker-Version: flag!
 
 // {{{ License, Contact, Notes & Includes 
 
@@ -32,7 +33,7 @@
 //   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 //
 //   Contact:
-//            
+//            Michael Brown <michaelb@opentext.com>
 //
 
 // Notes:
@@ -67,13 +68,17 @@
 
 #include <sys/types.h>
 #include <sys/wait.h>
+#include <sys/stat.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
 #include <string.h>
 #include <sysexits.h>
 #include <unistd.h>
+#include <fcntl.h>
 #include <syslog.h>
 #include <signal.h>
+#include <poll.h>
 #include <errno.h>
 
 // C++ includes
@@ -89,7 +94,7 @@ extern "C" {
 #endif
 
 #include "libmilter/mfapi.h"
-#include "libmilter/mfdef.h"
+//#include "libmilter/mfdef.h"
 
 #ifdef  __cplusplus
 }
@@ -99,34 +104,90 @@ extern "C" {
 
 // }}} 
 
+struct smfiDesc smfilter =
+  {
+    "SpamAssassin", // filter name
+    SMFI_VERSION,   // version code -- leave untouched
+    SMFIF_ADDHDRS|SMFIF_CHGHDRS|SMFIF_CHGBODY,  // flags
+    NULL, // info filter callback
+    NULL, // HELO filter callback
+    mlfi_envfrom, // envelope filter callback
+    NULL, // envelope recipient filter callback
+    mlfi_header, // header filter callback
+    mlfi_eoh, // end of header callback
+    mlfi_body, // body filter callback
+    mlfi_eom, // end of message callback
+    mlfi_abort, // message aborted callback
+    mlfi_close, // connection cleanup callback
+  };
+
+int flag_debug=0;
+
 // {{{ main()
 
 int
 main(int argc, char* argv[])
 {
-  // ONLY command line option: socket filename!
-  if ( argc != 2 )
-    {
-      cout << PACKAGE << "- Version" << VERSION << endl;
+   int c, err = 0;
+   const char *args = "p:fd:";
+   char *sock = NULL;
+   bool dofork = false;
+
+	/* Process command line options */
+	while ((c = getopt(argc, argv, args)) != -1) {
+		switch (c) {
+			case 'p':
+				sock = strdup(optarg);
+				break;
+			case 'f':
+				dofork = true;
+				break;
+			case 'd':
+				flag_debug = atoi(optarg);
+				break;
+			case '?':
+				err = 1;
+				break;
+		}
+	}
+
+   if (!sock || err) {
+      cout << PACKAGE_NAME << " - Version " << PACKAGE_VERSION << endl;
       cout << "SpamAssassin Sendmail Milter Plugin" << endl;
-      cout << "Usage: spamass-milter <socket>" << endl;
-      _exit(0);
-    };
+      cout << "Usage: spamass-milter -p socket [-f] [-d nn]" << endl;
+      cout << "   -p socket: path to create socket" << endl;
+      cout << "   -f: fork into background" << endl;
+      cout << "   -d nn: set debug level to nn (1 or 2).  Logs to syslog" << endl;
+      exit(EX_USAGE);
+   }
 
-  // set connection socket
-  smfi_setconn(argv[1]);
+	if (dofork == true) {
+		switch(fork()) {
+         case -1: /* Uh-oh, we have a problem forking. */
+            fprintf(stderr, "Uh-oh, couldn't fork!\n");
+				exit(errno);
+				break;
+			case 0: /* Child */
+				break;
+			default: /* Parent */
+				exit(0);
+		}
+	}
+   {
+      struct stat junk;
+      if (stat(sock,&junk) == 0) unlink(sock);
+   }
 
-  // register callbacks with Milter
-  if (smfi_register(smfilter) == MI_FAILURE)
-    {
-      throw_error("smfi_register failed");
-      cerr << "smfi_register failed. exit." << endl;
-      exit(EX_UNAVAILABLE);
-    };
-
-  // enter Milter main routine
-  return smfi_main();
+   openlog("spamass-milter", LOG_PID, LOG_MAIL);
 
+   (void) smfi_setconn(sock);
+	if (smfi_register(smfilter) == MI_FAILURE) {
+		fprintf(stderr, "smfi_register failed\n");
+		exit(EX_UNAVAILABLE);
+	} else {
+      debug(1, "smfi_register succeeded");
+   }
+	return smfi_main();
 };
 
 // }}}
@@ -204,6 +265,20 @@ assassinate(SMFICTX* ctx, SpamAssassin* assassin)
 		   const_cast<char*>(assassin->spam_prev_content_type().c_str()));
 
 
+  // X-Spam-Level header //
+  // find it:
+  old = assassin->set_spam_level(retrieve_field(assassin->d().substr(0, eoh), 
+						string("X-Spam-Level")));
+  
+  // change if old one was present, append if non-null
+  if (old > 0)
+    smfi_chgheader(ctx,"X-Spam-Level",1,assassin->spam_level().size() > 0 ? 
+		   const_cast<char*>(assassin->spam_level().c_str()) : NULL );
+  else if (assassin->spam_level().size()>0)
+    smfi_addheader(ctx, "X-Spam-Level", 
+		   const_cast<char*>(assassin->spam_level().c_str()));
+  
+
   // X-Spam-Checker-Version header //
   // find it:
   old = assassin->set_spam_checker_version(retrieve_field(assassin->d().substr(0, eoh), 
@@ -216,42 +291,52 @@ assassinate(SMFICTX* ctx, SpamAssassin* assassin)
   else if (assassin->spam_checker_version().size()>0)
     smfi_addheader(ctx, "X-Spam-Checker-Version", 
 		   const_cast<char*>(assassin->spam_checker_version().c_str()));
-  
+
+
   // 
   // If SpamAssassin thinks it is spam, replace
   //  Subject:
   //  Content-Type:
   //  <Body>
   // 
+  //  However, only issue the header replacement calls if the content has
+  //  actually changed. If SA didn't change subject or content-type, don't
+  //  replace here unnecessarily.
   if (assassin->spam_flag().size()>0)
     {
+      string oldstring;
+      string newstring;
 
       // Subject header //
       // find it:
-      old = assassin->set_subject(retrieve_field(assassin->d().substr(0, eoh), 
-						 string("Subject")));
+      newstring = retrieve_field(assassin->d().substr(0, eoh), string("Subject"));
+      oldstring = assassin->subject();
+
+      old = assassin->set_subject(oldstring);
       
       // change if old one was present, append if non-null
-      if (old > 0)
-	smfi_chgheader(ctx,"Subject",1,assassin->subject().size() > 0 ? 
-		       const_cast<char*>(assassin->subject().c_str()) : NULL );
-      else if (assassin->content_type().size()>0)
+      if (old > 0 && newstring != oldstring)
+	smfi_chgheader(ctx,"Subject",1,newstring.size() > 0 ? 
+		       const_cast<char*>(newstring.c_str()) : NULL );
+      else if (newstring.size()>0)
 	smfi_addheader(ctx, "Subject", 
-		       const_cast<char*>(assassin->subject().c_str()));
+		       const_cast<char*>(newstring.c_str()));
       
 
       // Content-Type header //
       // find it:
-      old = assassin->set_content_type(retrieve_field(assassin->d().substr(0, eoh), 
-						      string("Content-Type")));
+      newstring = retrieve_field(assassin->d().substr(0, eoh), string("Content-Type"));
+      oldstring = assassin->content_type();
+
+      old = assassin->set_content_type(oldstring);
       
       // change if old one was present, append if non-null
-      if (old > 0)
-	smfi_chgheader(ctx,"Content-Type",1,assassin->content_type().size() > 0 ? 
-		       const_cast<char*>(assassin->content_type().c_str()) : NULL );
+      if (old > 0 && newstring != oldstring)
+	smfi_chgheader(ctx,"Content-Type",1,newstring.size() > 0 ? 
+		       const_cast<char*>(newstring.c_str()) : NULL );
       else if (assassin->content_type().size()>0)
 	smfi_addheader(ctx, "Content-Type", 
-		       const_cast<char*>(assassin->content_type().c_str()));
+		       const_cast<char*>(newstring.c_str()));
       
       
       // Replace body with the one SpamAssassin provided //
@@ -312,6 +397,7 @@ mlfi_envfrom(SMFICTX* ctx, char** envfrom)
 {
   SpamAssassin* assassin;
 
+  debug(1, "mlfi_envfrom: enter");
   try {
     // launch new SpamAssassin
     assassin=new SpamAssassin;
@@ -325,6 +411,8 @@ mlfi_envfrom(SMFICTX* ctx, char** envfrom)
   smfi_setpriv(ctx, static_cast<void*>(assassin));
 
   // tell Milter to continue
+  debug(1, "mlfi_envfrom: exit");
+
   return SMFIS_CONTINUE;
 };
 
@@ -341,6 +429,7 @@ sfsistat
 mlfi_header(SMFICTX* ctx, char* headerf, char* headerv)
 {
   SpamAssassin* assassin = static_cast<SpamAssassin*>(smfi_getpriv(ctx));
+  debug(1, "mlfi_header: enter");
 
   // Is it a "X-Spam-" header field?
   if ( cmp_nocase_partial(string("X-Spam-"), string(headerf)) == 0 )
@@ -362,6 +451,10 @@ mlfi_header(SMFICTX* ctx, char* headerf, char* headerv)
       // X-Spam-Prev-Content-Type:
       if ( cmp_nocase_partial(string("X-Spam-Prev-Content-Type"), string(headerf)) == 0 )
 	assassin->set_spam_prev_content_type(string(headerv));
+
+      // X-Spam-Level:
+      if ( cmp_nocase_partial(string("X-Spam-Level"), string(headerf)) == 0 )
+	assassin->set_spam_level(string(headerv));
       
       // X-Spam-Checker-Version:
       if ( cmp_nocase_partial(string("X-Spam-Checker-Version"), string(headerf)) == 0 )
@@ -389,10 +482,13 @@ mlfi_header(SMFICTX* ctx, char* headerf, char* headerv)
       throw_error(problem);
       smfi_setpriv(ctx, static_cast<void*>(0));
       delete assassin;
+      debug(1, "mlfi_header: exit");
       return SMFIS_TEMPFAIL;
     };
   
   // go on...
+  debug(1, "mlfi_header: exit");
+
   return SMFIS_CONTINUE;
 };
 
@@ -407,18 +503,24 @@ mlfi_eoh(SMFICTX* ctx)
 {
   SpamAssassin* assassin = static_cast<SpamAssassin*>(smfi_getpriv(ctx));
 
+  debug(1, "mlfi_eoh: enter");
+
   try {
     // add blank line between header and body
-    assassin->output("\r\n",2);
+    assassin->output("\n\n",2);
   } catch (string& problem)
     {
       throw_error(problem);
       smfi_setpriv(ctx, static_cast<void*>(0));
       delete assassin;
+  
+      debug(1, "mlfi_eoh: exit");
       return SMFIS_TEMPFAIL;
     };
   
   // go on...
+
+  debug(1, "mlfi_eoh: exit");
   return SMFIS_CONTINUE;
 };
 
@@ -430,6 +532,7 @@ mlfi_eoh(SMFICTX* ctx)
 sfsistat
 mlfi_body(SMFICTX* ctx, u_char *bodyp, size_t bodylen)
 {
+  debug(1, "mlfi_body: enter");
   SpamAssassin* assassin = static_cast<SpamAssassin*>(smfi_getpriv(ctx));
  
   try {
@@ -439,10 +542,12 @@ mlfi_body(SMFICTX* ctx, u_char *bodyp, size_t bodylen)
       throw_error(problem);
       smfi_setpriv(ctx, static_cast<void*>(0));
       delete assassin;
+      debug(1, "mlfi_body: exit");
       return SMFIS_TEMPFAIL;
     };
 
   // go on...
+  debug(1, "mlfi_body: exit");
   return SMFIS_CONTINUE;
 };
 
@@ -458,6 +563,7 @@ mlfi_eom(SMFICTX* ctx)
 {
   SpamAssassin* assassin = static_cast<SpamAssassin*>(smfi_getpriv(ctx));
  
+  debug(1, "mlfi_eom: enter");
   try {
 
     // close output pipe to signal EOF to SpamAssassin
@@ -482,10 +588,12 @@ mlfi_eom(SMFICTX* ctx)
       throw_error(problem);
       smfi_setpriv(ctx, static_cast<void*>(0));
       delete assassin;
+      debug(1, "mlfi_eom: exit");
       return SMFIS_TEMPFAIL;
     };
   
   // go on...
+  debug(1, "mlfi_eom: exit");
   return SMFIS_CONTINUE;
 };
 
@@ -495,6 +603,7 @@ mlfi_eom(SMFICTX* ctx)
 sfsistat
 mlfi_close(SMFICTX* ctx)
 {
+  debug(1, "mlfi_close");
   return SMFIS_ACCEPT;
 };
 
@@ -509,6 +618,7 @@ mlfi_abort(SMFICTX* ctx)
 {
   SpamAssassin* assassin = static_cast<SpamAssassin*>(smfi_getpriv(ctx));
 
+  debug(1, "mlfi_abort");
   smfi_setpriv(ctx, static_cast<void*>(0));
   delete assassin;
 
@@ -549,12 +659,23 @@ SpamAssassin::SpamAssassin():
       dup2(pipe_io[1][1],1);
       dup2(pipe_io[1][1],2);
 
+      closeall(3);
+
       // execute spamc 
       // absolute path (determined in autoconf) 
       // should be a little more secure
-      char** argv = (char**) malloc(3*sizeof(char*));
+#if 0
+      char** argv = (char**) malloc(4*sizeof(char*));
+      argv[0] = SPAMC;
+      argv[1] = "-u";
+      argv[2] = "milter";
+      argv[3] = 0;
+#else
+      char** argv = (char**) malloc(2*sizeof(char*));
       argv[0] = SPAMC;
       argv[1] = 0;
+#endif
+
       execvp(argv[0] , argv); // does not return!
 
       // execution failed
@@ -571,6 +692,14 @@ SpamAssassin::SpamAssassin():
   pipe_io[0][0]=-1;
   pipe_io[1][1]=-1;
 
+  // mark the pipes non-blocking
+  if(fcntl(pipe_io[0][1], F_SETFL, O_NONBLOCK) == -1)
+     throw string(string("Cannot set pipe01 nonblocking: ")+string(strerror(errno)));
+#if 0  /* don't really need to make the sink pipe nonblocking */
+  if(fcntl(pipe_io[1][0], F_SETFL, O_NONBLOCK) == -1)
+     throw string(string("Cannot set pipe10 nonblocking: ")+string(strerror(errno)));
+#endif
+
   // we have to assume the client is running now.
   running=true;
 
@@ -609,44 +738,74 @@ SpamAssassin::~SpamAssassin()
 void
 SpamAssassin::output(const void* buffer, long size)
 {
+  debug(1, "::output enter");
+
   // if there are problems, fail.
   if (!running || error)
     throw string("tried output despite problems. failed.");
-  
+
   // send to SpamAssassin
   long total(0), wsize(0);
   string reason;
   int status;
   do {
-    switch(wsize=write(pipe_io[0][1], buffer, size-total))
-      {
-      case -1:
-	reason = string(strerror(errno));
+	struct pollfd fds[2];
+	int nfds = 2, nready;
+	fds[0].fd = pipe_io[0][1];
+	fds[0].events = POLLOUT;
+	fds[1].fd = pipe_io[1][0];
+	fds[1].events = POLLIN;
 
-	// close the pipes
-	close(pipe_io[0][1]);
-	close(pipe_io[1][0]);
-	pipe_io[0][1]=-1;	
-	pipe_io[1][0]=-1;	
+	debug(2, "polling");
+	nready = poll(fds, nfds, 1000);
+	if (nready == -1)
+		throw("poll failed");
 
-	// Slaughter child
-	kill(pid, SIGKILL);
+	debug(2, "poll returned %d", nready);
 
-	// set flags
-	error = true;
-	running = false;
+	if (fds[1].revents & POLLIN)
+	{
+		debug(2, "poll says I can read");
+		read_pipe();
+	}
+
+	if (fds[0].revents & POLLOUT)
+	{
+		debug(2, "poll says I can write");
+		switch(wsize = write(pipe_io[0][1], (char *)buffer + total, size - total))
+		{
+		  case -1:
+			if (errno == EAGAIN)
+				continue;
+			reason = string(strerror(errno));
+
+			// close the pipes
+			close(pipe_io[0][1]);
+			close(pipe_io[1][0]);
+			pipe_io[0][1]=-1;	
+			pipe_io[1][0]=-1;	
+
+			// Slaughter child
+			kill(pid, SIGKILL);
+
+			// set flags
+			error = true;
+			running = false;
 	
-	// wait until child is dead
-	waitpid(pid, &status, 0);
+			// wait until child is dead
+			waitpid(pid, &status, 0);
 
-	throw string(string("write error: ")+reason);	
-	break;
-      default:
-	total+=wsize;
-	break;
-      };
+			throw string(string("write error: ")+reason);	
+			break;
+	      default:
+			total += wsize;
+			debug(2, "wrote %d bytes");
+			break;
+		}
+	}
   } while ( total < size );
-  
+
+  debug(1, "::output exit");
 };
 
 // close output pipe
@@ -661,6 +820,7 @@ SpamAssassin::close_output()
 void
 SpamAssassin::input()
 {
+	debug(1, "::input enter");
   // if the child has exited or we experienced an error, return
   // immediately.
   if (!running || error)
@@ -679,6 +839,7 @@ SpamAssassin::input()
       error = true;
       throw string(string("waitpid error: ")+string(strerror(errno)));
     }; 
+	debug(1, "::input exit");
 };
 
 //
@@ -723,6 +884,12 @@ SpamAssassin::spam_checker_version()
   return x_spam_checker_version;
 };
 
+string& 
+SpamAssassin::spam_level()
+{
+  return x_spam_level;
+};
+
 string& 
 SpamAssassin::content_type()
 {
@@ -779,6 +946,14 @@ SpamAssassin::set_spam_checker_version(const string& val)
   return (old);
 };
 
+string::size_type
+SpamAssassin::set_spam_level(const string& val)
+{
+  string::size_type old = x_spam_level.size();
+  x_spam_level = val;
+  return (old);
+};
+
 string::size_type
 SpamAssassin::set_content_type(const string& val)
 {
@@ -795,6 +970,66 @@ SpamAssassin::set_subject(const string& val)
   return (old);
 };
 
+//
+// Read available output from SpamAssassin client
+//
+int
+SpamAssassin::read_pipe()
+{
+	long size;
+	int  status;
+	char iobuff[1024];
+	string reason;
+
+	debug(1, "::read_pipe enter");
+
+	if (pipe_io[1][0] == -1)
+	{
+		debug(1, "::read_pipe exit - shouldn't have been called?");
+		return 0;
+	}
+
+	size = read(pipe_io[1][0], iobuff, 1024);
+
+	if (size < 0)
+    {
+		// Error. 
+		reason = string(strerror(errno));
+		
+		// Close remaining pipe.
+		close(pipe_io[1][0]);
+		pipe_io[1][0] = -1;
+	
+		// Slaughter child
+		kill(pid, SIGKILL);
+	
+		// set flags
+		error = true;
+		running = false;
+	
+		// wait until child is dead
+		waitpid(pid, &status, 0);
+	
+		// throw the error message that caused this trouble
+		throw string(string("read error: ")+reason);
+	} else if ( size == 0 )
+	{
+
+		// EOF. Close the pipe
+		if(close(pipe_io[1][0]))
+			throw string(string("close error: ")+string(strerror(errno)));
+		pipe_io[1][0] = -1;
+	
+	} else
+	{ 
+		// append to mail buffer 
+		mail.append(iobuff, size);
+		debug(2, "read %d bytes");
+	}
+	debug(1, "::read_pipe exit");
+	return size;
+};
+
 //
 // Read all output from SpamAssassin client
 // and close the pipe
@@ -802,55 +1037,10 @@ SpamAssassin::set_subject(const string& val)
 void
 SpamAssassin::empty_and_close_pipe()
 {
-  long size;
-  char iobuff[1025];
-  int  status;
-  string reason;
-
-  do {
-
-    size=read(pipe_io[1][0],iobuff,1024);
-
-    if (size<0)
-      {
-
-	// Error. 
-	reason=string(strerror(errno));
-		
-	// Close remaining pipe.
-	close(pipe_io[1][0]);
-	pipe_io[1][0]=-1;
-	
-	// Slaughter child
-	kill(pid, SIGKILL);
-	
-	// set flags
-	error = true;
-	running = false;
-	
-	// wait until child is dead
-	waitpid(pid, &status, 0);
-	
-	// throw the error message that caused this trouble
-	throw string(string("read error: ")+reason);
-
-      }
-    else if ( size == 0 )
-      {
-
-	// EOF. Close the pipe
-	if(close(pipe_io[1][0]))
-	  throw string(string("close error: ")+string(strerror(errno)));
-	pipe_io[1][0]=-1;
-	
-      }
-    
-    // append to mail buffer 
-    iobuff[size]=0;
-    mail+=string(iobuff);
-    
-  } while (size > 0);
-
+	debug(1, "::empty_and_close_pipe enter");
+	while (read_pipe())
+		;
+	debug(1, "::empty_and_close_pipe exit");
 };
 
 // }}}
@@ -861,9 +1051,45 @@ SpamAssassin::empty_and_close_pipe()
 void
 throw_error(const string& errmsg)
 {
-  syslog(LOG_PERROR|LOG_PID|LOG_CONS|LOG_MAIL,errmsg.c_str());
+  syslog(LOG_ERR, errmsg.c_str());
 };
 
+void debug(int level, const char* string, ...)
+{
+	if (flag_debug >= level)
+	{
+#if defined(HAVE_VSYSLOG)
+	    va_list vl;
+	    va_start(vl, string);
+		vsyslog(LOG_ERR, string, vl);
+		va_end(vl);
+#else
+#if defined(HAVE_VASPRINTF)
+		char *buf;
+#else
+		char buf[1024];
+#endif
+	    va_list vl;
+	    va_start(vl, string);
+#if defined(HAVE_VASPRINTF)
+	    vasprintf(&buf, string, vl);
+#else
+#if defined(HAVE_VSNPRINTF)
+	    vsnprintf(buf, sizeof(buf)-1, string, vl);
+#else
+		/* XXX possible buffer overflow here; be careful what you pass to debug() */
+		vsprintf(buf, string, vl);
+#endif
+#endif
+	    va_end(vl);
+		syslog(LOG_ERR, "%s", buf);
+#if defined(HAVE_VASPRINTF)
+	    free(buf);
+#endif 
+#endif /* vsyslog */
+	}
+}
+
 // case-insensitive search 
 string::size_type 
 find_nocase(const string& array, const string& pattern, string::size_type start)
@@ -906,4 +1132,13 @@ cmp_nocase_partial(const string& s, const string& s2)
 
 };
 
+/* closeall() - close all FDs >= a specified value */ 
+void closeall(int fd) 
+{
+	int fdlimit = sysconf(_SC_OPEN_MAX); 
+	while (fd < fdlimit) 
+		close(fd++); 
+}
+
 // }}}
+// vim6:ai:noexpandtab
