@@ -1,10 +1,32 @@
 //-*-c++-*-
 //
-//  $Id: spamass-milter.h,v 1.2 2002/01/31 15:28:50 greve Exp $
+//  $Id: spamass-milter.h,v 1.5 2002/07/23 02:02:12 dnelson Exp $
 //
 //  Main include file for SpamAss-Milter
 //
+//  Copyright (c) 2002 Georg C. F. Greve <greve@gnu.org>,
+//   all rights maintained by FSF Europe e.V., 
+//   Villa Vogelsang, Antonienallee 1, 45279 Essen, Germany
+//
+//   This program is free software; you can redistribute it and/or modify
+//   it under the terms of the GNU General Public License as published by
+//   the Free Software Foundation; either version 2 of the License, or
+//   (at your option) any later version.
+//  
+//   This program is distributed in the hope that it will be useful,
+//   but WITHOUT ANY WARRANTY; without even the implied warranty of
+//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+//   GNU General Public License for more details.
+//  
+//   You should have received a copy of the GNU General Public License
+//   along with this program; if not, write to the Free Software
+//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+//
+//   Contact:
+//            Michael Brown <michaelb@opentext.com>
+//
 
+using namespace std;
 
 string retrieve_field(const string&, const string&);
 
@@ -17,22 +39,7 @@ sfsistat mlfi_close(SMFICTX*);
 sfsistat mlfi_abort(SMFICTX*);
 sfsistat mlfi_abort(SMFICTX*);
 
-struct smfiDesc smfilter =
-  {
-    "SpamAssassin", // filter name
-    SMFI_VERSION,   // version code -- leave untouched
-    SMFIF_ADDHDRS|SMFIF_CHGHDRS|SMFIF_CHGBODY,  // flags
-    NULL, // info filter callback
-    NULL, // HELO filter callback
-    mlfi_envfrom, // envelope filter callback
-    NULL, // envelope recipient filter callback
-    mlfi_header, // header filter callback
-    mlfi_eoh, // end of header callback
-    mlfi_body, // body filter callback
-    mlfi_eom, // end of message callback
-    mlfi_abort, // message aborted callback
-    mlfi_close, // connection cleanup callback
-  };
+extern struct smfiDesc smfilter;
 
 class SpamAssassin {
 public:
@@ -50,6 +57,7 @@ public:
   string& spam_report();
   string& spam_prev_content_type();
   string& spam_checker_version();
+  string& spam_level();
   string& content_type();
   string& subject();
   string::size_type set_spam_status(const string&);
@@ -57,11 +65,13 @@ public:
   string::size_type set_spam_report(const string&);
   string::size_type set_spam_prev_content_type(const string&);
   string::size_type set_spam_checker_version(const string&);
+  string::size_type set_spam_level(const string&);
   string::size_type set_content_type(const string&);
   string::size_type set_subject(const string&);
 
 private:
   void empty_and_close_pipe();
+  int read_pipe();
 
 public:  
   // flags
@@ -74,7 +84,7 @@ public:
 
   // Variables for SpamAssassin influenced fields
   string x_spam_status, x_spam_flag, x_spam_report, x_spam_prev_content_type;
-  string x_spam_checker_version, _content_type, _subject;
+  string x_spam_checker_version, x_spam_level, _content_type, _subject;
 
   // Process handling variables
   pid_t pid;
@@ -85,6 +95,7 @@ public:
 void assassinate(SMFICTX*, SpamAssassin*);
 
 void throw_error(const string&);
+void debug(int level, const char* string, ...);
 string::size_type find_nocase(const string&, const string&, string::size_type = 0);
 int cmp_nocase_partial(const string&, const string&);
-
+void closeall(int fd);
